# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Database Migration Service conversion workspaces CRUD API."""

from typing import Any, Mapping, Optional

from googlecloudsdk.api_lib.database_migration import api_util
from googlecloudsdk.api_lib.database_migration.conversion_workspaces import base_conversion_workspaces_client
from googlecloudsdk.api_lib.database_migration.conversion_workspaces import conversion_workspace_builder
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.database_migration.conversion_workspaces import enums
from googlecloudsdk.generated_clients.apis.datamigration.v1 import datamigration_v1_messages as messages


class ConversionWorkspacesCRUDClient(
    base_conversion_workspaces_client.BaseConversionWorkspacesClient
):
  """Client for Conversion Workspaces CRUD API."""

  @property
  def cw_builder(
      self,
  ) -> conversion_workspace_builder.ConversionWorkspaceBuilder:
    """Returns an instance of the conversion workspace builder."""
    return conversion_workspace_builder.ConversionWorkspaceBuilder(
        release_track=self.release_track,
    )

  def Create(
      self,
      parent_ref: str,
      conversion_workspace_id: str,
      display_name: str,
      source_database_provider: enums.SourceDatabaseProvider,
      source_database_engine: enums.SourceDatabaseEngine,
      source_database_version: str,
      destination_database_provider: enums.DestinationDatabaseProvider,
      destination_database_engine: enums.DestinationDatabaseEngine,
      destination_database_version: str,
      global_settings: messages.ConversionWorkspace.GlobalSettingsValue,
  ) -> messages.Operation:
    """Creates a conversion workspace.

    Args:
      parent_ref: a Resource reference to a parent
        datamigration.projects.locations resource for this conversion workspace.
      conversion_workspace_id: the name of the resource to create.
      display_name: the display name for the conversion workspace.
      source_database_provider: the source database provider for the conversion
        workspace.
      source_database_engine: the source database engine for the conversion
        workspace.
      source_database_version: the source database version for the conversion
        workspace.
      destination_database_provider: the destination database provider for the
        conversion workspace.
      destination_database_engine: the destination database engine for the
        conversion workspace.
      destination_database_version: the destination database version for the
        conversion workspace.
      global_settings: GlobalSettings, the global settings for the conversion
        workspace.

    Returns:
      Operation: the operation for creating the conversion workspace.
    """
    if global_settings and global_settings.additionalProperties:
      props = {
          prop.key: prop.value for prop in global_settings.additionalProperties
      }
      if (
          props.get('autoAssessment') == 'true'
          and props.get('aiConversion') == 'false'
      ):
        raise exceptions.InvalidArgumentException(
            '--quality-assessment',
            'cannot be true if --auto-conversion is false.',
        )
    return self.cw_service.Create(
        self.messages.DatamigrationProjectsLocationsConversionWorkspacesCreateRequest(
            conversionWorkspace=self.cw_builder.Build(
                display_name=display_name,
                source_database_provider=source_database_provider,
                source_database_engine=source_database_engine,
                source_database_version=source_database_version,
                destination_database_provider=destination_database_provider,
                destination_database_engine=destination_database_engine,
                destination_database_version=destination_database_version,
                global_settings=global_settings,
            ),
            conversionWorkspaceId=conversion_workspace_id,
            parent=parent_ref,
            requestId=api_util.GenerateRequestId(),
        ),
    )

  def Read(self, name: str):
    return self.cw_service.Get(
        self.messages.DatamigrationProjectsLocationsConversionWorkspacesGetRequest(
            name=name,
        )
    )

  def GetGlobalFilter(self, name: str) -> str:
    """Get global filter for a conversion workspace.

    If no global filter is set, '*' will be returned.

    Args:
      name: The name of the conversion workspace.

    Returns:
      The global filter for the conversion workspace.
    """
    return self._GetAdditionalProperties(name).get('filter', '*')

  def _GetAdditionalProperties(self, name: str) -> Mapping[str, Any]:
    """Get conversion workspace additional properties.

    Args:
      name: The name of the conversion workspace.

    Returns:
      The conversion workspace additional properties.
    """
    conversion_workspace = self.Read(name=name)
    if not conversion_workspace.globalSettings:
      return {}

    return {
        additional_property.key: additional_property.value
        for additional_property in (
            conversion_workspace.globalSettings.additionalProperties
        )
    }

  def Update(
      self,
      name: str,
      display_name: Optional[str],
      source_database_name_override: Optional[str],
      global_filter: Optional[str],
      auto_conversion: Optional[bool] = None,
      quality_assessment: Optional[bool] = None,
      conversion_assistance: Optional[bool] = None,
      pattern_matching: Optional[bool] = None,
  ):
    """Updates a conversion workspace.

    Args:
      name: str, the reference of the conversion workspace to update.
      display_name: the display name to update.
      source_database_name_override: the source database name override to
        update.
      global_filter: the global filter for the conversion workspace.
      auto_conversion: whether to enable auto-conversion.
      quality_assessment: whether to enable quality assessment.
      conversion_assistance: whether to enable conversion assistance.
      pattern_matching: whether to enable pattern matching.

    Returns:
      Operation: the operation for updating the conversion workspace.
    """
    conversion_workspace, update_fields = self._GetUpdatedConversionWorkspace(
        conversion_workspace=self.Read(name),
        display_name=display_name,
        source_database_name_override=source_database_name_override,
        global_filter=global_filter,
        auto_conversion=auto_conversion,
        quality_assessment=quality_assessment,
        conversion_assistance=conversion_assistance,
        pattern_matching=pattern_matching,
    )
    return self.cw_service.Patch(
        self.messages.DatamigrationProjectsLocationsConversionWorkspacesPatchRequest(
            conversionWorkspace=conversion_workspace,
            name=name,
            requestId=api_util.GenerateRequestId(),
            updateMask=','.join(update_fields),
        ),
    )

  def Delete(self, name: str):
    """Deletes a conversion workspace.

    Args:
      name: str, the name of the resource to delete.

    Returns:
      Operation: the operation for deleting the conversion workspace.
    """
    return self.cw_service.Delete(
        self.messages.DatamigrationProjectsLocationsConversionWorkspacesDeleteRequest(
            name=name,
            requestId=api_util.GenerateRequestId(),
        ),
    )

  def _GetUpdatedConversionWorkspace(
      self,
      conversion_workspace: messages.ConversionWorkspace,
      display_name: Optional[str],
      source_database_name_override: Optional[str],
      global_filter: Optional[str],
      auto_conversion: Optional[bool] = None,
      quality_assessment: Optional[bool] = None,
      conversion_assistance: Optional[bool] = None,
      pattern_matching: Optional[bool] = None,
  ) -> tuple[messages.ConversionWorkspace, set[str]]:
    """Returns updated conversion workspace and list of updated fields.

    Args:
      conversion_workspace: the conversion workspace to update.
      display_name: the display name to update.
      source_database_name_override: the source database name override to
        update.
      global_filter: the global filter for the conversion workspace.
      auto_conversion: whether to enable auto-conversion.
      quality_assessment: whether to enable quality assessment.
      conversion_assistance: whether to enable conversion assistance.
      pattern_matching: whether to enable pattern matching.

    Returns:
      A tuple containing:
        The updated conversion workspace object.
        A set of the names of the updated fields.
    """
    update_fields = set()

    if display_name:
      conversion_workspace.displayName = display_name
      update_fields.add('displayName')

    current_settings = {}
    if conversion_workspace.globalSettings:
      for prop in (
          conversion_workspace.globalSettings.additionalProperties or []
      ):
        current_settings[prop.key] = prop.value

    new_settings = current_settings.copy()

    if source_database_name_override is not None:
      new_settings['sourceDatabaseOverride'] = source_database_name_override

    if auto_conversion is not None:
      new_settings['aiConversion'] = str(auto_conversion).lower()
      if not auto_conversion:
        new_settings['autoAssessment'] = str(auto_conversion).lower()
    if quality_assessment is not None:
      new_settings['autoAssessment'] = str(quality_assessment).lower()
    if global_filter is not None:
      new_settings['filter'] = global_filter
    if conversion_assistance is not None:
      new_settings['aiAssistant'] = str(conversion_assistance).lower()
    if pattern_matching is not None:
      new_settings['aiPatternApplier'] = str(pattern_matching).lower()

    if (
        new_settings.get('autoAssessment') == 'true'
        and new_settings.get('aiConversion', 'false') == 'false'
    ):
      raise exceptions.InvalidArgumentException(
          '--quality-assessment',
          'cannot be true if auto-conversion is false.',
      )

    if new_settings != current_settings:
      update_fields.add('globalSettings')
      properties = []
      for key, value in sorted(new_settings.items()):
        properties.append(
            self.messages.ConversionWorkspace.GlobalSettingsValue.AdditionalProperty(
                key=key, value=value
            )
        )
      conversion_workspace.globalSettings = (
          self.messages.ConversionWorkspace.GlobalSettingsValue(
              additionalProperties=properties
          )
      )

    return conversion_workspace, update_fields



"""Generated message classes for iamconnectors version v1alpha.

"""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.protorpclite import messages as _messages
from apitools.base.py import encoding
from apitools.base.py import extra_types


package = 'iamconnectors'


class AccessEvent(_messages.Message):
  r"""Message describing AccessEvent object

  Messages:
    LabelsValue: Optional. Labels as key value pairs

  Fields:
    connector: Output only. The connector that this access event is associated
      with.
    firstAccessTime: Output only. The first time this user has interacted with
      this agent. Rounded to the previous hour.
    labels: Optional. Labels as key value pairs
    lastAccessTime: Output only. The most recent time this user has interacted
      with this agent. Rounded to the previous hour.
    name: Output only. Identifier. Name of the AccessEvent
    purgeTime: Output only. The time when this access event is permanently
      deleted.
    scopes: Output only. All scopes that have been used by this user with this
      agent. The number of scopes is limited to 200.
    tokenEndpoint: Output only. The url of the authentication server that was
      accessed.
    userId: Output only. The user_id provided by the agent application for
      this user. Not verified by Google.
    workload: Output only. The identity bound to the workload that this user
      interacted with to produce this AccessEvent. Will typically be an
      agentic spiffe id
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Optional. Labels as key value pairs

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  connector = _messages.StringField(1)
  firstAccessTime = _messages.StringField(2)
  labels = _messages.MessageField('LabelsValue', 3)
  lastAccessTime = _messages.StringField(4)
  name = _messages.StringField(5)
  purgeTime = _messages.StringField(6)
  scopes = _messages.StringField(7, repeated=True)
  tokenEndpoint = _messages.StringField(8)
  userId = _messages.StringField(9)
  workload = _messages.StringField(10)


class ApiKeyParams(_messages.Message):
  r"""Message describing ApiKeyParams object.

  Fields:
    apiKey: Optional. Input only. The API key for this connector.
  """

  apiKey = _messages.StringField(1)


class AuditConfig(_messages.Message):
  r"""Specifies the audit configuration for a service. The configuration
  determines which permission types are logged, and what identities, if any,
  are exempted from logging. An AuditConfig must have one or more
  AuditLogConfigs. If there are AuditConfigs for both `allServices` and a
  specific service, the union of the two AuditConfigs is used for that
  service: the log_types specified in each AuditConfig are enabled, and the
  exempted_members in each AuditLogConfig are exempted. Example Policy with
  multiple AuditConfigs: { "audit_configs": [ { "service": "allServices",
  "audit_log_configs": [ { "log_type": "DATA_READ", "exempted_members": [
  "user:jose@example.com" ] }, { "log_type": "DATA_WRITE" }, { "log_type":
  "ADMIN_READ" } ] }, { "service": "sampleservice.googleapis.com",
  "audit_log_configs": [ { "log_type": "DATA_READ" }, { "log_type":
  "DATA_WRITE", "exempted_members": [ "user:aliya@example.com" ] } ] } ] } For
  sampleservice, this policy enables DATA_READ, DATA_WRITE and ADMIN_READ
  logging. It also exempts `jose@example.com` from DATA_READ logging, and
  `aliya@example.com` from DATA_WRITE logging.

  Fields:
    auditLogConfigs: The configuration for logging of each type of permission.
    service: Specifies a service that will be enabled for audit logging. For
      example, `storage.googleapis.com`, `cloudsql.googleapis.com`.
      `allServices` is a special value that covers all services.
  """

  auditLogConfigs = _messages.MessageField('AuditLogConfig', 1, repeated=True)
  service = _messages.StringField(2)


class AuditLogConfig(_messages.Message):
  r"""Provides the configuration for logging a type of permissions. Example: {
  "audit_log_configs": [ { "log_type": "DATA_READ", "exempted_members": [
  "user:jose@example.com" ] }, { "log_type": "DATA_WRITE" } ] } This enables
  'DATA_READ' and 'DATA_WRITE' logging, while exempting jose@example.com from
  DATA_READ logging.

  Enums:
    LogTypeValueValuesEnum: The log type that this config enables.

  Fields:
    exemptedMembers: Specifies the identities that do not cause logging for
      this type of permission. Follows the same format of Binding.members.
    logType: The log type that this config enables.
  """

  class LogTypeValueValuesEnum(_messages.Enum):
    r"""The log type that this config enables.

    Values:
      LOG_TYPE_UNSPECIFIED: Default case. Should never be this.
      ADMIN_READ: Admin reads. Example: CloudIAM getIamPolicy
      DATA_WRITE: Data writes. Example: CloudSQL Users create
      DATA_READ: Data reads. Example: CloudSQL Users list
    """
    LOG_TYPE_UNSPECIFIED = 0
    ADMIN_READ = 1
    DATA_WRITE = 2
    DATA_READ = 3

  exemptedMembers = _messages.StringField(1, repeated=True)
  logType = _messages.EnumField('LogTypeValueValuesEnum', 2)


class Authorization(_messages.Message):
  r"""Message describing Authorization object

  Enums:
    StateValueValuesEnum: Output only. The state of the authorization.

  Fields:
    clientUserId: Output only. The client_user_id provided by the client
      application for their end user. Not verified by Google.
    createTime: Output only. [Output only] Create time stamp
    name: Identifier. name of resource
    scopes: Output only. The scopes actually granted by the end user during
      the consent flow.
    state: Output only. The state of the authorization.
    updateTime: Output only. [Output only] Update time stamp
  """

  class StateValueValuesEnum(_messages.Enum):
    r"""Output only. The state of the authorization.

    Values:
      STATE_UNSPECIFIED: Unspecified state.
      ACTIVE: Active.
      SUSPENDED: Suspended.
    """
    STATE_UNSPECIFIED = 0
    ACTIVE = 1
    SUSPENDED = 2

  clientUserId = _messages.StringField(1)
  createTime = _messages.StringField(2)
  name = _messages.StringField(3)
  scopes = _messages.StringField(4, repeated=True)
  state = _messages.EnumField('StateValueValuesEnum', 5)
  updateTime = _messages.StringField(6)


class Binding(_messages.Message):
  r"""Associates `members`, or principals, with a `role`.

  Fields:
    condition: The condition that is associated with this binding. If the
      condition evaluates to `true`, then this binding applies to the current
      request. If the condition evaluates to `false`, then this binding does
      not apply to the current request. However, a different role binding
      might grant the same role to one or more of the principals in this
      binding. To learn which resources support conditions in their IAM
      policies, see the [IAM
      documentation](https://cloud.google.com/iam/help/conditions/resource-
      policies).
    members: Specifies the principals requesting access for a Google Cloud
      resource. `members` can have the following values: * `allUsers`: A
      special identifier that represents anyone who is on the internet; with
      or without a Google account. * `allAuthenticatedUsers`: A special
      identifier that represents anyone who is authenticated with a Google
      account or a service account. Does not include identities that come from
      external identity providers (IdPs) through identity federation. *
      `user:{emailid}`: An email address that represents a specific Google
      account. For example, `alice@example.com` . *
      `serviceAccount:{emailid}`: An email address that represents a Google
      service account. For example, `my-other-
      app@appspot.gserviceaccount.com`. *
      `serviceAccount:{projectid}.svc.id.goog[{namespace}/{kubernetes-sa}]`:
      An identifier for a [Kubernetes service
      account](https://cloud.google.com/kubernetes-engine/docs/how-
      to/kubernetes-service-accounts). For example, `my-
      project.svc.id.goog[my-namespace/my-kubernetes-sa]`. *
      `group:{emailid}`: An email address that represents a Google group. For
      example, `admins@example.com`. * `domain:{domain}`: The G Suite domain
      (primary) that represents all the users of that domain. For example,
      `google.com` or `example.com`. * `principal://iam.googleapis.com/locatio
      ns/global/workforcePools/{pool_id}/subject/{subject_attribute_value}`: A
      single identity in a workforce identity pool. * `principalSet://iam.goog
      leapis.com/locations/global/workforcePools/{pool_id}/group/{group_id}`:
      All workforce identities in a group. * `principalSet://iam.googleapis.co
      m/locations/global/workforcePools/{pool_id}/attribute.{attribute_name}/{
      attribute_value}`: All workforce identities with a specific attribute
      value. * `principalSet://iam.googleapis.com/locations/global/workforcePo
      ols/{pool_id}/*`: All identities in a workforce identity pool. * `princi
      pal://iam.googleapis.com/projects/{project_number}/locations/global/work
      loadIdentityPools/{pool_id}/subject/{subject_attribute_value}`: A single
      identity in a workload identity pool. * `principalSet://iam.googleapis.c
      om/projects/{project_number}/locations/global/workloadIdentityPools/{poo
      l_id}/group/{group_id}`: A workload identity pool group. * `principalSet
      ://iam.googleapis.com/projects/{project_number}/locations/global/workloa
      dIdentityPools/{pool_id}/attribute.{attribute_name}/{attribute_value}`:
      All identities in a workload identity pool with a certain attribute. * `
      principalSet://iam.googleapis.com/projects/{project_number}/locations/gl
      obal/workloadIdentityPools/{pool_id}/*`: All identities in a workload
      identity pool. * `deleted:user:{emailid}?uid={uniqueid}`: An email
      address (plus unique identifier) representing a user that has been
      recently deleted. For example,
      `alice@example.com?uid=123456789012345678901`. If the user is recovered,
      this value reverts to `user:{emailid}` and the recovered user retains
      the role in the binding. *
      `deleted:serviceAccount:{emailid}?uid={uniqueid}`: An email address
      (plus unique identifier) representing a service account that has been
      recently deleted. For example, `my-other-
      app@appspot.gserviceaccount.com?uid=123456789012345678901`. If the
      service account is undeleted, this value reverts to
      `serviceAccount:{emailid}` and the undeleted service account retains the
      role in the binding. * `deleted:group:{emailid}?uid={uniqueid}`: An
      email address (plus unique identifier) representing a Google group that
      has been recently deleted. For example,
      `admins@example.com?uid=123456789012345678901`. If the group is
      recovered, this value reverts to `group:{emailid}` and the recovered
      group retains the role in the binding. * `deleted:principal://iam.google
      apis.com/locations/global/workforcePools/{pool_id}/subject/{subject_attr
      ibute_value}`: Deleted single identity in a workforce identity pool. For
      example, `deleted:principal://iam.googleapis.com/locations/global/workfo
      rcePools/my-pool-id/subject/my-subject-attribute-value`.
    role: Role that is assigned to the list of `members`, or principals. For
      example, `roles/viewer`, `roles/editor`, or `roles/owner`. For an
      overview of the IAM roles and permissions, see the [IAM
      documentation](https://cloud.google.com/iam/docs/roles-overview). For a
      list of the available pre-defined roles, see
      [here](https://cloud.google.com/iam/docs/understanding-roles).
  """

  condition = _messages.MessageField('Expr', 1)
  members = _messages.StringField(2, repeated=True)
  role = _messages.StringField(3)


class CancelOperationRequest(_messages.Message):
  r"""The request message for Operations.CancelOperation."""


class Connector(_messages.Message):
  r"""Message describing Connector object

  Enums:
    StateValueValuesEnum: Optional. The state of the connector.

  Messages:
    LabelsValue: Optional. Labels as key value pairs

  Fields:
    allowedScopes: Optional. List of scopes that are allowed to be requested
      for this connector. If this list is non-empty, only scopes within this
      list may be requested. If this list is empty, all scopes may be
      requested. Scopes appearing in `blocked_scopes` are disallowed even if
      they appear in `allowed_scopes`. The number of allowed scopes is limited
      to 200.
    blockedScopes: Optional. List of scopes that are blocked from being
      requested for this connector. If a scope appears in this list, it may
      not be requested, even if it also appears in `allowed_scopes`.
      `blocked_scopes` takes precedence over `allowed_scopes`. The number of
      blocked scopes is limited to 200.
    connectorTypeParams: Required. Connector type specific parameters.
    createTime: Output only. [Output only] Create time stamp
    deleted: Output only. This is set to true if the connector is deleted.
    description: Optional. Description of the resource. Must be less than 256
      characters.
    expireTime: Output only. The time when the connector will expire.
    labels: Optional. Labels as key value pairs
    name: Identifier. The full resource name of the connector. Format:
      projects/{project}/locations/{location}/connectors/{connector}
    state: Optional. The state of the connector.
    updateTime: Output only. [Output only] Update time stamp
    workloadIds: Optional. Input only. Workload identity in IAM principal://
      format of the agent(s) that will use this connector. Example: `principal
      ://agents.global.org-
      ${ORG_ID}.system.id.goog/resources/aiplatform/projects/{PROJECT_ID}/loca
      tions/{LOCATIONS}/reasoningEngines/{ID}`
  """

  class StateValueValuesEnum(_messages.Enum):
    r"""Optional. The state of the connector.

    Values:
      STATE_UNSPECIFIED: Unspecified state.
      ENABLED: Enabled and can be used.
      DISABLED: Disabled and cannot be used.
    """
    STATE_UNSPECIFIED = 0
    ENABLED = 1
    DISABLED = 2

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Optional. Labels as key value pairs

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  allowedScopes = _messages.StringField(1, repeated=True)
  blockedScopes = _messages.StringField(2, repeated=True)
  connectorTypeParams = _messages.MessageField('ConnectorTypeParams', 3)
  createTime = _messages.StringField(4)
  deleted = _messages.BooleanField(5)
  description = _messages.StringField(6)
  expireTime = _messages.StringField(7)
  labels = _messages.MessageField('LabelsValue', 8)
  name = _messages.StringField(9)
  state = _messages.EnumField('StateValueValuesEnum', 10)
  updateTime = _messages.StringField(11)
  workloadIds = _messages.StringField(12, repeated=True)


class ConnectorTypeParams(_messages.Message):
  r"""Connector type specific parameters. Required when creating a connector.

  Fields:
    apiKey: ApiKey connector type parameters.
    geConnectorParams: GeminiEnterprise connector type parameters.
    threeLeggedOauth: ThreeLeggedOAuth connector type parameters.
    twoLeggedOauth: TwoLeggedOAuth connector type parameters.
  """

  apiKey = _messages.MessageField('ApiKeyParams', 1)
  geConnectorParams = _messages.MessageField('GeminiEnterpriseConnectorParams', 2)
  threeLeggedOauth = _messages.MessageField('ThreeLeggedOAuth', 3)
  twoLeggedOauth = _messages.MessageField('TwoLeggedOAuth', 4)


class Empty(_messages.Message):
  r"""A generic empty message that you can re-use to avoid defining duplicated
  empty messages in your APIs. A typical example is to use it as the request
  or the response type of an API method. For instance: service Foo { rpc
  Bar(google.protobuf.Empty) returns (google.protobuf.Empty); }
  """



class Expr(_messages.Message):
  r"""Represents a textual expression in the Common Expression Language (CEL)
  syntax. CEL is a C-like expression language. The syntax and semantics of CEL
  are documented at https://github.com/google/cel-spec. Example (Comparison):
  title: "Summary size limit" description: "Determines if a summary is less
  than 100 chars" expression: "document.summary.size() < 100" Example
  (Equality): title: "Requestor is owner" description: "Determines if
  requestor is the document owner" expression: "document.owner ==
  request.auth.claims.email" Example (Logic): title: "Public documents"
  description: "Determine whether the document should be publicly visible"
  expression: "document.type != 'private' && document.type != 'internal'"
  Example (Data Manipulation): title: "Notification string" description:
  "Create a notification string with a timestamp." expression: "'New message
  received at ' + string(document.create_time)" The exact variables and
  functions that may be referenced within an expression are determined by the
  service that evaluates it. See the service documentation for additional
  information.

  Fields:
    description: Optional. Description of the expression. This is a longer
      text which describes the expression, e.g. when hovered over it in a UI.
    expression: Textual representation of an expression in Common Expression
      Language syntax.
    location: Optional. String indicating the location of the expression for
      error reporting, e.g. a file name and a position in the file.
    title: Optional. Title for the expression, i.e. a short string describing
      its purpose. This can be used e.g. in UIs which allow to enter the
      expression.
  """

  description = _messages.StringField(1)
  expression = _messages.StringField(2)
  location = _messages.StringField(3)
  title = _messages.StringField(4)


class GeminiEnterpriseConnectorParams(_messages.Message):
  r"""Message describing GeminiEnterpriseConnectorParams object."""


class IamconnectorsProjectsLocationsAccessEventsGetRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsAccessEventsGetRequest object.

  Fields:
    name: Required. Name of the resource
  """

  name = _messages.StringField(1, required=True)


class IamconnectorsProjectsLocationsAccessEventsListRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsAccessEventsListRequest object.

  Fields:
    filter: Optional. Filter string to restrict the results. Currently
      supports filtering by `workload_id` or `connector_name`. If no filter is
      provided, returns all access events for the requested project and
      location. Format: `workload_id=""` or `connector_name=""`
    orderBy: Optional. This field is currently ignored. Defaults to ordering
      by (connector_id, user_id) in ascending order.
    pageSize: Optional. Requested page size. Server may return fewer items
      than requested. If unspecified, server will pick an appropriate default.
    pageToken: Optional. A token identifying a page of results the server
      should return.
    parent: Required. The parent resource where the search is performed.
      Format: projects/{project}/locations/{location}
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class IamconnectorsProjectsLocationsConnectorsAuthorizationsDeleteRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsAuthorizationsDeleteRequest
  object.

  Fields:
    name: Required. The name of the Authorization to delete. Format: projects/
      {project}/locations/{location}/connectors/{connector}/authorizations/{au
      thorization}
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes after the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
  """

  name = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)


class IamconnectorsProjectsLocationsConnectorsAuthorizationsGetRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsAuthorizationsGetRequest
  object.

  Fields:
    name: Required. Name of the resource
  """

  name = _messages.StringField(1, required=True)


class IamconnectorsProjectsLocationsConnectorsAuthorizationsListRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsAuthorizationsListRequest
  object.

  Fields:
    filter: Optional. Filter string to restrict the results. Currently
      supports filtering by `client_user_id` only. Format: `client_user_id=""`
    orderBy: Optional. This field is currently ignored. Defaults to ordering
      by authorization_id in ascending order.
    pageSize: Optional. Requested page size. Server may return fewer items
      than requested. If unspecified, server will pick an appropriate default.
    pageToken: Optional. A page token, received from a previous
      `ListAuthorizations` call. Provide this to retrieve the subsequent page.
      When paginating, all other parameters provided to `ListAuthorizations`
      must match the call that provided the page token.
    parent: Required. The parent resource where the search is performed.
      Format: projects/{project}/locations/{location}/connectors/{connector}
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class IamconnectorsProjectsLocationsConnectorsCreateRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsCreateRequest object.

  Fields:
    connector: A Connector resource to be passed as the request body.
    connectorId: Required. The ID to use for the connector, which will become
      the final segment of the connector's resource name. This value should be
      1-63 characters, and valid characters are /a-z-/. The first character
      must be a lowercase letter, and the last character must be a lowercase
      letter or a number.
    parent: Required. The parent resource where the connector is created.
      Format: projects/{project}/locations/{location}
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes since the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
  """

  connector = _messages.MessageField('Connector', 1)
  connectorId = _messages.StringField(2)
  parent = _messages.StringField(3, required=True)
  requestId = _messages.StringField(4)


class IamconnectorsProjectsLocationsConnectorsDeleteRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsDeleteRequest object.

  Fields:
    name: Required. Name of the resource
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes after the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
  """

  name = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)


class IamconnectorsProjectsLocationsConnectorsGetIamPolicyRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsGetIamPolicyRequest object.

  Fields:
    options_requestedPolicyVersion: Optional. The maximum policy version that
      will be used to format the policy. Valid values are 0, 1, and 3.
      Requests specifying an invalid value will be rejected. Requests for
      policies with any conditional role bindings must specify version 3.
      Policies with no conditional role bindings may specify any valid value
      or leave the field unset. The policy in the response might use the
      policy version that you specified, or it might use a lower policy
      version. For example, if you specify version 3, but the policy has no
      conditional role bindings, the response uses version 1. To learn which
      resources support conditions in their IAM policies, see the [IAM
      documentation](https://cloud.google.com/iam/help/conditions/resource-
      policies).
    resource: REQUIRED: The resource for which the policy is being requested.
      See [Resource
      names](https://cloud.google.com/apis/design/resource_names) for the
      appropriate value for this field.
  """

  options_requestedPolicyVersion = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  resource = _messages.StringField(2, required=True)


class IamconnectorsProjectsLocationsConnectorsGetRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsGetRequest object.

  Fields:
    name: Required. Name of the resource
  """

  name = _messages.StringField(1, required=True)


class IamconnectorsProjectsLocationsConnectorsListRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsListRequest object.

  Fields:
    filter: Optional. Filter results. This field is currently ignored.
    orderBy: Optional. Currently ignored. Defaults to ordering by connector_id
      in ascending order.
    pageSize: Optional. Requested page size. Server may return fewer items
      than requested. If unspecified, server will pick an appropriate default.
    pageToken: Optional. A token, which can be sent as `page_token` to
      retrieve the next page. If this field is omitted, the first page is
      returned.
    parent: Required. The parent resource where the search is performed.
      Format: projects/{project}/locations/{location}
    showDeleted: Optional. Deleted connectors will be kept with a soft-delete
      for 30 days before being purged. If this field is set to true, deleted
      connectors will also be returned.
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)
  showDeleted = _messages.BooleanField(6)


class IamconnectorsProjectsLocationsConnectorsPatchRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsPatchRequest object.

  Fields:
    connector: A Connector resource to be passed as the request body.
    name: Identifier. The full resource name of the connector. Format:
      projects/{project}/locations/{location}/connectors/{connector}
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes since the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
    updateMask: Optional. Field mask is used to specify the fields to be
      overwritten in the Connector resource by the update. The fields
      specified in the update_mask are relative to the resource, not the full
      request. A field will be overwritten if it is in the mask. If the user
      does not provide a mask then all fields present in the request will be
      overwritten.
  """

  connector = _messages.MessageField('Connector', 1)
  name = _messages.StringField(2, required=True)
  requestId = _messages.StringField(3)
  updateMask = _messages.StringField(4)


class IamconnectorsProjectsLocationsConnectorsQueryRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsQueryRequest object.

  Fields:
    pageSize: Optional. Requested page size. Server may return fewer items
      than requested. If unspecified, server will pick an appropriate default.
      The maximum page size is 1000.
    pageToken: Optional. A token, which can be sent as `page_token` to
      retrieve the next page. If this field is omitted, the first page is
      returned. A page token, received from a previous QueryConnectors call.
      Provide this to retrieve the subsequent page. When paginating, all other
      parameters provided to QueryConnectors must match the call that provided
      the page token.
    parent: Required. The parent resource where the search is performed.
      Format: projects/{project}/locations/{location}
    workloadId: Required. The workload identifier to filter by.
  """

  pageSize = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(2)
  parent = _messages.StringField(3, required=True)
  workloadId = _messages.StringField(4)


class IamconnectorsProjectsLocationsConnectorsQueryWorkloadsRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsQueryWorkloadsRequest object.

  Fields:
    name: Required. The name of the connector to query. Format:
      projects/{project}/locations/{location}/connectors/{connector}
    pageSize: Optional. Requested page size. Server may return fewer items
      than requested. If unspecified, server will pick an appropriate default.
    pageToken: Optional. A token, which can be sent as `page_token` to
      retrieve the next page. When paginating, all other parameters provided
      to QueryWorkloads must match the call that provided the page token. If
      this field is omitted, the first page is returned.
  """

  name = _messages.StringField(1, required=True)
  pageSize = _messages.IntegerField(2, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(3)


class IamconnectorsProjectsLocationsConnectorsRevokeAuthorizationRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsRevokeAuthorizationRequest
  object.

  Fields:
    name: Required. The resource name of the Connector. Format:
      projects/{project}/locations/{location}/connectors/{connector}
    revokeAuthorizationRequest: A RevokeAuthorizationRequest resource to be
      passed as the request body.
  """

  name = _messages.StringField(1, required=True)
  revokeAuthorizationRequest = _messages.MessageField('RevokeAuthorizationRequest', 2)


class IamconnectorsProjectsLocationsConnectorsSetIamPolicyRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsSetIamPolicyRequest object.

  Fields:
    resource: REQUIRED: The resource for which the policy is being specified.
      See [Resource
      names](https://cloud.google.com/apis/design/resource_names) for the
      appropriate value for this field.
    setIamPolicyRequest: A SetIamPolicyRequest resource to be passed as the
      request body.
  """

  resource = _messages.StringField(1, required=True)
  setIamPolicyRequest = _messages.MessageField('SetIamPolicyRequest', 2)


class IamconnectorsProjectsLocationsConnectorsTestIamPermissionsRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsTestIamPermissionsRequest
  object.

  Fields:
    resource: REQUIRED: The resource for which the policy detail is being
      requested. See [Resource
      names](https://cloud.google.com/apis/design/resource_names) for the
      appropriate value for this field.
    testIamPermissionsRequest: A TestIamPermissionsRequest resource to be
      passed as the request body.
  """

  resource = _messages.StringField(1, required=True)
  testIamPermissionsRequest = _messages.MessageField('TestIamPermissionsRequest', 2)


class IamconnectorsProjectsLocationsConnectorsUndeleteRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsConnectorsUndeleteRequest object.

  Fields:
    name: Required. Name of the resource Format:
      projects/{project}/locations/{location}/connectors/{connector}
    undeleteConnectorRequest: A UndeleteConnectorRequest resource to be passed
      as the request body.
  """

  name = _messages.StringField(1, required=True)
  undeleteConnectorRequest = _messages.MessageField('UndeleteConnectorRequest', 2)


class IamconnectorsProjectsLocationsGetRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsGetRequest object.

  Fields:
    name: Resource name for the location.
  """

  name = _messages.StringField(1, required=True)


class IamconnectorsProjectsLocationsListRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsListRequest object.

  Fields:
    extraLocationTypes: Optional. Do not use this field unless explicitly
      documented otherwise. This is primarily for internal usage.
    filter: A filter to narrow down results to a preferred subset. The
      filtering language accepts strings like `"displayName=tokyo"`, and is
      documented in more detail in [AIP-160](https://google.aip.dev/160).
    name: The resource that owns the locations collection, if applicable.
    pageSize: The maximum number of results to return. If not set, the service
      selects a default.
    pageToken: A page token received from the `next_page_token` field in the
      response. Send that page token to receive the subsequent page.
  """

  extraLocationTypes = _messages.StringField(1, repeated=True)
  filter = _messages.StringField(2)
  name = _messages.StringField(3, required=True)
  pageSize = _messages.IntegerField(4, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(5)


class IamconnectorsProjectsLocationsOperationsCancelRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsOperationsCancelRequest object.

  Fields:
    cancelOperationRequest: A CancelOperationRequest resource to be passed as
      the request body.
    name: The name of the operation resource to be cancelled.
  """

  cancelOperationRequest = _messages.MessageField('CancelOperationRequest', 1)
  name = _messages.StringField(2, required=True)


class IamconnectorsProjectsLocationsOperationsDeleteRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsOperationsDeleteRequest object.

  Fields:
    name: The name of the operation resource to be deleted.
  """

  name = _messages.StringField(1, required=True)


class IamconnectorsProjectsLocationsOperationsGetRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsOperationsGetRequest object.

  Fields:
    name: The name of the operation resource.
  """

  name = _messages.StringField(1, required=True)


class IamconnectorsProjectsLocationsOperationsListRequest(_messages.Message):
  r"""A IamconnectorsProjectsLocationsOperationsListRequest object.

  Fields:
    filter: The standard list filter.
    name: The name of the operation's parent resource.
    pageSize: The standard list page size.
    pageToken: The standard list page token.
    returnPartialSuccess: When set to `true`, operations that are reachable
      are returned as normal, and those that are unreachable are returned in
      the ListOperationsResponse.unreachable field. This can only be `true`
      when reading across collections. For example, when `parent` is set to
      `"projects/example/locations/-"`. This field is not supported by default
      and will result in an `UNIMPLEMENTED` error if set unless explicitly
      documented otherwise in service or product specific documentation.
  """

  filter = _messages.StringField(1)
  name = _messages.StringField(2, required=True)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  returnPartialSuccess = _messages.BooleanField(5)


class ListAccessEventsResponse(_messages.Message):
  r"""Message for response to listing AccessEvents

  Fields:
    accessEvents: The list of AccessEvent
    nextPageToken: A token identifying a page of results the server should
      return.
    unreachable: Unordered list. Locations that could not be reached.
  """

  accessEvents = _messages.MessageField('AccessEvent', 1, repeated=True)
  nextPageToken = _messages.StringField(2)
  unreachable = _messages.StringField(3, repeated=True)


class ListAuthorizationsResponse(_messages.Message):
  r"""Message for response to listing Authorizations

  Fields:
    authorizations: The list of Authorization
    nextPageToken: A token identifying a page of results the server should
      return.
    unreachable: Unordered list. Locations that could not be reached.
  """

  authorizations = _messages.MessageField('Authorization', 1, repeated=True)
  nextPageToken = _messages.StringField(2)
  unreachable = _messages.StringField(3, repeated=True)


class ListConnectorsResponse(_messages.Message):
  r"""Message for response to listing Connectors

  Fields:
    connectors: The list of Connector
    nextPageToken: A token identifying a page of results the server should
      return.
    unreachable: Unordered list. Locations that could not be reached.
  """

  connectors = _messages.MessageField('Connector', 1, repeated=True)
  nextPageToken = _messages.StringField(2)
  unreachable = _messages.StringField(3, repeated=True)


class ListLocationsResponse(_messages.Message):
  r"""The response message for Locations.ListLocations.

  Fields:
    locations: A list of locations that matches the specified filter in the
      request.
    nextPageToken: The standard List next-page token.
  """

  locations = _messages.MessageField('Location', 1, repeated=True)
  nextPageToken = _messages.StringField(2)


class ListOperationsResponse(_messages.Message):
  r"""The response message for Operations.ListOperations.

  Fields:
    nextPageToken: The standard List next-page token.
    operations: A list of operations that matches the specified filter in the
      request.
    unreachable: Unordered list. Unreachable resources. Populated when the
      request sets `ListOperationsRequest.return_partial_success` and reads
      across collections. For example, when attempting to list all resources
      across all supported locations.
  """

  nextPageToken = _messages.StringField(1)
  operations = _messages.MessageField('Operation', 2, repeated=True)
  unreachable = _messages.StringField(3, repeated=True)


class Location(_messages.Message):
  r"""A resource that represents a Google Cloud location.

  Messages:
    LabelsValue: Cross-service attributes for the location. For example
      {"cloud.googleapis.com/region": "us-east1"}
    MetadataValue: Service-specific metadata. For example the available
      capacity at the given location.

  Fields:
    displayName: The friendly name for this location, typically a nearby city
      name. For example, "Tokyo".
    labels: Cross-service attributes for the location. For example
      {"cloud.googleapis.com/region": "us-east1"}
    locationId: The canonical id for this location. For example: `"us-east1"`.
    metadata: Service-specific metadata. For example the available capacity at
      the given location.
    name: Resource name for the location, which may vary between
      implementations. For example: `"projects/example-project/locations/us-
      east1"`
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Cross-service attributes for the location. For example
    {"cloud.googleapis.com/region": "us-east1"}

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class MetadataValue(_messages.Message):
    r"""Service-specific metadata. For example the available capacity at the
    given location.

    Messages:
      AdditionalProperty: An additional property for a MetadataValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a MetadataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  displayName = _messages.StringField(1)
  labels = _messages.MessageField('LabelsValue', 2)
  locationId = _messages.StringField(3)
  metadata = _messages.MessageField('MetadataValue', 4)
  name = _messages.StringField(5)


class Operation(_messages.Message):
  r"""This resource represents a long-running operation that is the result of
  a network API call.

  Messages:
    MetadataValue: Service-specific metadata associated with the operation. It
      typically contains progress information and common metadata such as
      create time. Some services might not provide such metadata. Any method
      that returns a long-running operation should document the metadata type,
      if any.
    ResponseValue: The normal, successful response of the operation. If the
      original method returns no data on success, such as `Delete`, the
      response is `google.protobuf.Empty`. If the original method is standard
      `Get`/`Create`/`Update`, the response should be the resource. For other
      methods, the response should have the type `XxxResponse`, where `Xxx` is
      the original method name. For example, if the original method name is
      `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.

  Fields:
    done: If the value is `false`, it means the operation is still in
      progress. If `true`, the operation is completed, and either `error` or
      `response` is available.
    error: The error result of the operation in case of failure or
      cancellation.
    metadata: Service-specific metadata associated with the operation. It
      typically contains progress information and common metadata such as
      create time. Some services might not provide such metadata. Any method
      that returns a long-running operation should document the metadata type,
      if any.
    name: The server-assigned name, which is only unique within the same
      service that originally returns it. If you use the default HTTP mapping,
      the `name` should be a resource name ending with
      `operations/{unique_id}`.
    response: The normal, successful response of the operation. If the
      original method returns no data on success, such as `Delete`, the
      response is `google.protobuf.Empty`. If the original method is standard
      `Get`/`Create`/`Update`, the response should be the resource. For other
      methods, the response should have the type `XxxResponse`, where `Xxx` is
      the original method name. For example, if the original method name is
      `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class MetadataValue(_messages.Message):
    r"""Service-specific metadata associated with the operation. It typically
    contains progress information and common metadata such as create time.
    Some services might not provide such metadata. Any method that returns a
    long-running operation should document the metadata type, if any.

    Messages:
      AdditionalProperty: An additional property for a MetadataValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a MetadataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ResponseValue(_messages.Message):
    r"""The normal, successful response of the operation. If the original
    method returns no data on success, such as `Delete`, the response is
    `google.protobuf.Empty`. If the original method is standard
    `Get`/`Create`/`Update`, the response should be the resource. For other
    methods, the response should have the type `XxxResponse`, where `Xxx` is
    the original method name. For example, if the original method name is
    `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.

    Messages:
      AdditionalProperty: An additional property for a ResponseValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a ResponseValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  done = _messages.BooleanField(1)
  error = _messages.MessageField('Status', 2)
  metadata = _messages.MessageField('MetadataValue', 3)
  name = _messages.StringField(4)
  response = _messages.MessageField('ResponseValue', 5)


class OperationMetadata(_messages.Message):
  r"""Represents the metadata of the long-running operation.

  Fields:
    apiVersion: Output only. API version used to start the operation.
    createTime: Output only. The time the operation was created.
    endTime: Output only. The time the operation finished running.
    requestedCancellation: Output only. Identifies whether the user has
      requested cancellation of the operation. Operations that have been
      cancelled successfully have google.longrunning.Operation.error value
      with a google.rpc.Status.code of `1`, corresponding to `Code.CANCELLED`.
    statusMessage: Output only. Human-readable status of the operation, if
      any.
    target: Output only. Server-defined resource path for the target of the
      operation.
    verb: Output only. Name of the verb executed by the operation.
  """

  apiVersion = _messages.StringField(1)
  createTime = _messages.StringField(2)
  endTime = _messages.StringField(3)
  requestedCancellation = _messages.BooleanField(4)
  statusMessage = _messages.StringField(5)
  target = _messages.StringField(6)
  verb = _messages.StringField(7)


class Policy(_messages.Message):
  r"""An Identity and Access Management (IAM) policy, which specifies access
  controls for Google Cloud resources. A `Policy` is a collection of
  `bindings`. A `binding` binds one or more `members`, or principals, to a
  single `role`. Principals can be user accounts, service accounts, Google
  groups, and domains (such as G Suite). A `role` is a named list of
  permissions; each `role` can be an IAM predefined role or a user-created
  custom role. For some types of Google Cloud resources, a `binding` can also
  specify a `condition`, which is a logical expression that allows access to a
  resource only if the expression evaluates to `true`. A condition can add
  constraints based on attributes of the request, the resource, or both. To
  learn which resources support conditions in their IAM policies, see the [IAM
  documentation](https://cloud.google.com/iam/help/conditions/resource-
  policies). **JSON example:** ``` { "bindings": [ { "role":
  "roles/resourcemanager.organizationAdmin", "members": [
  "user:mike@example.com", "group:admins@example.com", "domain:google.com",
  "serviceAccount:my-project-id@appspot.gserviceaccount.com" ] }, { "role":
  "roles/resourcemanager.organizationViewer", "members": [
  "user:eve@example.com" ], "condition": { "title": "expirable access",
  "description": "Does not grant access after Sep 2020", "expression":
  "request.time < timestamp('2020-10-01T00:00:00.000Z')", } } ], "etag":
  "BwWWja0YfJA=", "version": 3 } ``` **YAML example:** ``` bindings: -
  members: - user:mike@example.com - group:admins@example.com -
  domain:google.com - serviceAccount:my-project-id@appspot.gserviceaccount.com
  role: roles/resourcemanager.organizationAdmin - members: -
  user:eve@example.com role: roles/resourcemanager.organizationViewer
  condition: title: expirable access description: Does not grant access after
  Sep 2020 expression: request.time < timestamp('2020-10-01T00:00:00.000Z')
  etag: BwWWja0YfJA= version: 3 ``` For a description of IAM and its features,
  see the [IAM documentation](https://cloud.google.com/iam/docs/).

  Fields:
    auditConfigs: Specifies cloud audit logging configuration for this policy.
    bindings: Associates a list of `members`, or principals, with a `role`.
      Optionally, may specify a `condition` that determines how and when the
      `bindings` are applied. Each of the `bindings` must contain at least one
      principal. The `bindings` in a `Policy` can refer to up to 1,500
      principals; up to 250 of these principals can be Google groups. Each
      occurrence of a principal counts towards these limits. For example, if
      the `bindings` grant 50 different roles to `user:alice@example.com`, and
      not to any other principal, then you can add another 1,450 principals to
      the `bindings` in the `Policy`.
    etag: `etag` is used for optimistic concurrency control as a way to help
      prevent simultaneous updates of a policy from overwriting each other. It
      is strongly suggested that systems make use of the `etag` in the read-
      modify-write cycle to perform policy updates in order to avoid race
      conditions: An `etag` is returned in the response to `getIamPolicy`, and
      systems are expected to put that etag in the request to `setIamPolicy`
      to ensure that their change will be applied to the same version of the
      policy. **Important:** If you use IAM Conditions, you must include the
      `etag` field whenever you call `setIamPolicy`. If you omit this field,
      then IAM allows you to overwrite a version `3` policy with a version `1`
      policy, and all of the conditions in the version `3` policy are lost.
    version: Specifies the format of the policy. Valid values are `0`, `1`,
      and `3`. Requests that specify an invalid value are rejected. Any
      operation that affects conditional role bindings must specify version
      `3`. This requirement applies to the following operations: * Getting a
      policy that includes a conditional role binding * Adding a conditional
      role binding to a policy * Changing a conditional role binding in a
      policy * Removing any role binding, with or without a condition, from a
      policy that includes conditions **Important:** If you use IAM
      Conditions, you must include the `etag` field whenever you call
      `setIamPolicy`. If you omit this field, then IAM allows you to overwrite
      a version `3` policy with a version `1` policy, and all of the
      conditions in the version `3` policy are lost. If a policy does not
      include any conditions, operations on that policy may specify any valid
      version or leave the field unset. To learn which resources support
      conditions in their IAM policies, see the [IAM
      documentation](https://cloud.google.com/iam/help/conditions/resource-
      policies).
  """

  auditConfigs = _messages.MessageField('AuditConfig', 1, repeated=True)
  bindings = _messages.MessageField('Binding', 2, repeated=True)
  etag = _messages.BytesField(3)
  version = _messages.IntegerField(4, variant=_messages.Variant.INT32)


class QueryConnectorsResponse(_messages.Message):
  r"""Response message for QueryConnectors.

  Fields:
    connectorNames: The unique list of connector resource names used by the
      workload.
    nextPageToken: A token identifying a page of results the server should
      return. If this field is omitted, there are no subsequent pages.
  """

  connectorNames = _messages.StringField(1, repeated=True)
  nextPageToken = _messages.StringField(2)


class QueryWorkloadsResponse(_messages.Message):
  r"""Response message for QueryWorkloads.

  Fields:
    nextPageToken: A token to retrieve the next page of results.
    workloadIds: The unique list of workload identifiers (agents) that used
      the connector.
  """

  nextPageToken = _messages.StringField(1)
  workloadIds = _messages.StringField(2, repeated=True)


class RevokeAuthorizationRequest(_messages.Message):
  r"""Request message for RevokeAuthorization.

  Fields:
    userId: Required. The identity of the user to revoke authorization for.
  """

  userId = _messages.StringField(1)


class RevokeAuthorizationResponse(_messages.Message):
  r"""Response message for RevokeAuthorization."""


class SetIamPolicyRequest(_messages.Message):
  r"""Request message for `SetIamPolicy` method.

  Fields:
    policy: REQUIRED: The complete policy to be applied to the `resource`. The
      size of the policy is limited to a few 10s of KB. An empty policy is a
      valid policy but certain Google Cloud services (such as Projects) might
      reject them.
    updateMask: OPTIONAL: A FieldMask specifying which fields of the policy to
      modify. Only the fields in the mask will be modified. If no mask is
      provided, the following default mask is used: `paths: "bindings, etag"`
  """

  policy = _messages.MessageField('Policy', 1)
  updateMask = _messages.StringField(2)


class StandardQueryParameters(_messages.Message):
  r"""Query parameters accepted by all methods.

  Enums:
    FXgafvValueValuesEnum: V1 error format.
    AltValueValuesEnum: Data format for response.

  Fields:
    f__xgafv: V1 error format.
    access_token: OAuth access token.
    alt: Data format for response.
    callback: JSONP
    fields: Selector specifying which fields to include in a partial response.
    key: API key. Your API key identifies your project and provides you with
      API access, quota, and reports. Required unless you provide an OAuth 2.0
      token.
    oauth_token: OAuth 2.0 token for the current user.
    prettyPrint: Returns response with indentations and line breaks.
    quotaUser: Available to use for quota purposes for server-side
      applications. Can be any arbitrary string assigned to a user, but should
      not exceed 40 characters.
    trace: A tracing token of the form "token:<tokenid>" to include in api
      requests.
    uploadType: Legacy upload protocol for media (e.g. "media", "multipart").
    upload_protocol: Upload protocol for media (e.g. "raw", "multipart").
  """

  class AltValueValuesEnum(_messages.Enum):
    r"""Data format for response.

    Values:
      json: Responses with Content-Type of application/json
      media: Media download with context-dependent Content-Type
      proto: Responses with Content-Type of application/x-protobuf
    """
    json = 0
    media = 1
    proto = 2

  class FXgafvValueValuesEnum(_messages.Enum):
    r"""V1 error format.

    Values:
      _1: v1 error format
      _2: v2 error format
    """
    _1 = 0
    _2 = 1

  f__xgafv = _messages.EnumField('FXgafvValueValuesEnum', 1)
  access_token = _messages.StringField(2)
  alt = _messages.EnumField('AltValueValuesEnum', 3, default='json')
  callback = _messages.StringField(4)
  fields = _messages.StringField(5)
  key = _messages.StringField(6)
  oauth_token = _messages.StringField(7)
  prettyPrint = _messages.BooleanField(8, default=True)
  quotaUser = _messages.StringField(9)
  trace = _messages.StringField(10)
  uploadType = _messages.StringField(11)
  upload_protocol = _messages.StringField(12)


class Status(_messages.Message):
  r"""The `Status` type defines a logical error model that is suitable for
  different programming environments, including REST APIs and RPC APIs. It is
  used by [gRPC](https://github.com/grpc). Each `Status` message contains
  three pieces of data: error code, error message, and error details. You can
  find out more about this error model and how to work with it in the [API
  Design Guide](https://cloud.google.com/apis/design/errors).

  Messages:
    DetailsValueListEntry: A DetailsValueListEntry object.

  Fields:
    code: The status code, which should be an enum value of google.rpc.Code.
    details: A list of messages that carry the error details. There is a
      common set of message types for APIs to use.
    message: A developer-facing error message, which should be in English. Any
      user-facing error message should be localized and sent in the
      google.rpc.Status.details field, or localized by the client.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class DetailsValueListEntry(_messages.Message):
    r"""A DetailsValueListEntry object.

    Messages:
      AdditionalProperty: An additional property for a DetailsValueListEntry
        object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a DetailsValueListEntry object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  code = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  details = _messages.MessageField('DetailsValueListEntry', 2, repeated=True)
  message = _messages.StringField(3)


class TestIamPermissionsRequest(_messages.Message):
  r"""Request message for `TestIamPermissions` method.

  Fields:
    permissions: The set of permissions to check for the `resource`.
      Permissions with wildcards (such as `*` or `storage.*`) are not allowed.
      For more information see [IAM
      Overview](https://cloud.google.com/iam/docs/overview#permissions).
  """

  permissions = _messages.StringField(1, repeated=True)


class TestIamPermissionsResponse(_messages.Message):
  r"""Response message for `TestIamPermissions` method.

  Fields:
    permissions: A subset of `TestPermissionsRequest.permissions` that the
      caller is allowed.
  """

  permissions = _messages.StringField(1, repeated=True)


class ThreeLeggedOAuth(_messages.Message):
  r"""Message describing ThreeLeggedOAuth object.

  Fields:
    authorizationUrl: Optional. The authorization endpoint to send users to
      for consenting to delegate to the agent. eg.
      "https://auth.atlassian.com/authorize"
    clientId: Optional. The client ID of the OAuth client.
    clientSecret: Optional. Input only. The client secret of the OAuth client.
    defaultContinueUri: Optional. The default continue URI for 3LO flow and it
      will be used when no continue URI is provided in the RetrieveCredentials
      request.
    enablePkce: Optional. Enables Proof Key for Code Exchange (PKCE) for the
      OAuth flow to prevent authorization code interception attacks.
    redirectUrl: Output only. The redirect URL this connector uses for the
      OAuth exchange. This is deterministic based on the name of the
      connector.
    tokenUrl: Optional. The token endpoint for requesting tokens on behalf of
      an end user. eg. "https://auth.atlassian.com/oauth/token"
  """

  authorizationUrl = _messages.StringField(1)
  clientId = _messages.StringField(2)
  clientSecret = _messages.StringField(3)
  defaultContinueUri = _messages.StringField(4)
  enablePkce = _messages.BooleanField(5)
  redirectUrl = _messages.StringField(6)
  tokenUrl = _messages.StringField(7)


class TwoLeggedOAuth(_messages.Message):
  r"""Message describing TwoLeggedOAuth object.

  Fields:
    clientId: Optional. The client ID of the OAuth client.
    clientSecret: Optional. Input only. The client secret of the OAuth client.
    tokenEndpoint: Optional. The token endpoint of the OAuth client.
  """

  clientId = _messages.StringField(1)
  clientSecret = _messages.StringField(2)
  tokenEndpoint = _messages.StringField(3)


class UndeleteConnectorRequest(_messages.Message):
  r"""Message for undeleting a Connector

  Fields:
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes after the first
      request. The request ID must be a valid UUID with the exception that
      zero UUID is not supported (00000000-0000-0000-0000-000000000000).
  """

  requestId = _messages.StringField(1)


encoding.AddCustomJsonFieldMapping(
    StandardQueryParameters, 'f__xgafv', '$.xgafv')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_1', '1')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_2', '2')
encoding.AddCustomJsonFieldMapping(
    IamconnectorsProjectsLocationsConnectorsGetIamPolicyRequest, 'options_requestedPolicyVersion', 'options.requestedPolicyVersion')



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to create App Lifecycle Manager Flags from a manifest file."""

from __future__ import annotations

import json
import pprint
import re
from typing import Any, Callable
from apitools.base.protorpclite import messages as rpclite_messages
from apitools.base.py import encoding
from apitools.base.py import exceptions
import googlecloudsdk.api_lib.saasservicemgmt.util as saasservicemgmt_util
from googlecloudsdk.api_lib.util import messages as msg_utils
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.command_lib.saas_runtime import flags as arg_utils
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.generated_clients.apis.saasservicemgmt.v1beta1 import saasservicemgmt_v1beta1_client as ezclient
from googlecloudsdk.generated_clients.apis.saasservicemgmt.v1beta1 import saasservicemgmt_v1beta1_messages as ezmessages


FLAG_TYPE_MAP: dict[str, str] = {
    'boolean': 'FLAG_VALUE_TYPE_BOOLEAN',
    'integer': 'FLAG_VALUE_TYPE_INTEGER',
    'double': 'FLAG_VALUE_TYPE_DOUBLE',
    'string': 'FLAG_VALUE_TYPE_STRING',
}

VALUE_TYPE_HANDLERS: dict[str, tuple[str, type[Any]]] = {
    'FLAG_VALUE_TYPE_BOOLEAN': ('booleanValue', bool),
    'FLAG_VALUE_TYPE_INTEGER': ('integerValue', int),
    'FLAG_VALUE_TYPE_DOUBLE': ('doubleValue', float),
    'FLAG_VALUE_TYPE_STRING': ('stringValue', str),
}

_NON_ALPHANUM_PATTERN = re.compile(r'[^a-z0-9]+')
_LEADING_NON_CHAR_PATTERN = re.compile(r'^[^a-z]+')


def _ValidFlagType(flag_type: str | None) -> bool:
  """Returns True if the flag type is valid."""
  return flag_type is not None and flag_type in FLAG_TYPE_MAP


def _ParseFlagData(
    flag_key: str,
    unit_kind_name: str,
    flag_definition: dict[str, Any],
) -> dict[str, Any]:
  """Parses a flag definition and returns a dict of fields to be set."""
  log.debug(f'Flag definition:\n{pprint.pformat(flag_definition)}\n')

  flag_type = flag_definition.get('flagType')
  if not _ValidFlagType(flag_type):
    raise exceptions.InvalidDataError(
        f'Flag "{flag_key}" invalid. "{flag_type}" is not a supported flag'
        ' type.'
    )

  value_type = FLAG_TYPE_MAP.get(flag_type)

  default_value = flag_definition.get('defaultValue')
  if default_value is None:
    raise exceptions.InvalidDataError(
        f'Flag "{flag_key}" invalid. "defaultValue" is not set.'
    )

  default_evaluation_spec, default_variant = (
      _GetDefaultEvalSpecAndVariantsDicts(default_value, value_type)
  )

  flag_dict = {
      'key': flag_key,
      'flagValueType': value_type,
      'evaluationSpec': default_evaluation_spec,
      'variants': [default_variant],
      'unitKind': unit_kind_name,
  }
  if flag_definition.get('description'):
    flag_dict['description'] = flag_definition['description']

  return flag_dict


def _GetDefaultEvalSpecAndVariantsDicts(
    default_value: str, value_type: str
) -> tuple[dict[str, Any], dict[str, Any]]:
  """Returns the default evaluation spec and default variant for the given value type."""

  if value_type not in VALUE_TYPE_HANDLERS:
    raise exceptions.InvalidDataError(f'Unsupported value type: {value_type}')

  variant_type, expected_type = VALUE_TYPE_HANDLERS[
      value_type
  ]
  if not isinstance(default_value, expected_type):
    raise exceptions.InvalidDataError(
        f'Invalid default value "{default_value}" for type {value_type}.'
        f' Expected type {expected_type.__name__} but got'
        f' {type(default_value).__name__}.'
    )

  variant_dict = {
      'id': 'default',
  }
  variant_dict[variant_type] = expected_type(default_value)

  return {'defaultTarget': 'default'}, variant_dict


def _GetFlagMessagesFromManifest(
    manifest_json: dict[str, Any], unit_kind_name: str
) -> list[ezmessages.Flag]:
  """Parses a JSON manifest file and returns a list of flags."""

  if 'flags' not in manifest_json:
    raise exceptions.InvalidDataError(
        'Manifest must have a top-level "flags" key.'
    )

  flags_data = manifest_json['flags']
  if not isinstance(flags_data, dict):
    raise exceptions.InvalidDataError('"flags" property is invalid.')

  has_error = False
  flags = []
  for flag_key, flag_definition in flags_data.items():
    try:
      flag_dict = _ParseFlagData(flag_key, unit_kind_name, flag_definition)
      flag_msg = msg_utils.DictToMessageWithErrorCheck(
          flag_dict, ezmessages.Flag, throw_on_unexpected_fields=False
      )
      flags.append(flag_msg)
    except (
        rpclite_messages.ValidationError,
        exceptions.InvalidDataError,
    ) as e:
      log.error(f'Failed to parse flag definition for flag "{flag_key}": {e!r}')
      has_error = True

  if has_error:
    raise exceptions.InvalidDataError(
        'Failed to parse some flags in the manifest.'
    )

  return flags


def _GetValidFlags(
    flags: list[ezmessages.Flag],
    flags_service: ezclient.SaasservicemgmtV1beta1.ProjectsLocationsFlagsService,
    parent: str,
) -> list[ezmessages.Flag]:
  """Returns a list of flags that are valid to create."""
  valid_flags = []
  validation_failed = False
  for flag_msg in flags:
    try:
      _CreateFlag(flags_service, parent, flag_msg, validate_only=True)
    except exceptions.HttpConflictError:
      log.status.Print(f'Flag {flag_msg.key} already exists. Ignoring.')
    except exceptions.HttpError as e:
      _LogHttpError(e, flag_msg.key, log.error, 'Flag creation would fail')
      validation_failed = True
    except exceptions.InvalidDataError as e:
      log.error(f'Flag creation would fail: {e}')
      validation_failed = True
    else:
      valid_flags.append(flag_msg)

  if validation_failed:
    raise exceptions.InvalidDataError('Manifest file contains invalid flags')

  return valid_flags


def _IsFlagAlreadyExistsError(e: exceptions.HttpError) -> bool:
  """Returns True if the error is due to a flag already existing."""
  if not isinstance(e, exceptions.HttpError):
    return False

  if e.status_code != 400:
    return False

  try:
    error_content = json.loads(e.content)
    error_message = error_content.get('error', {}).get('message', '')
    return 'already exists' in error_message
  except json.JSONDecodeError:
    return False


def _GetExistingFlag(
    flags_service: ezclient.SaasservicemgmtV1beta1.ProjectsLocationsFlagsService,
    parent: str,
    flag_key: str,
) -> ezmessages.Flag | None:
  """Gets a flag by its key."""
  try:
    get_request = ezmessages.SaasservicemgmtProjectsLocationsFlagsGetRequest(
        name=f'{parent}/flags/{flag_key}'
    )
    return flags_service.Get(get_request)
  except exceptions.HttpError as e:
    _LogHttpError(e, flag_key, log.error, 'Failed to get existing flag')
    return None


def _SanitizeFlagId(flag_key: str) -> str:
  """Sanitizes a flag key into a valid Google Cloud resource ID.

  Converts the provided flag key into a string that adheres to Google Cloud
  resource ID naming conventions, specifically matching the regular expression:
  ^[a-z]([-a-z0-9]{0,61}[a-z0-9])?$

  Args:
    flag_key: The original flag key from the manifest file.

  Returns:
    A sanitized string valid for use as a Google Cloud resource ID.

  Raises:
    exceptions.InvalidDataError: If the flag key contains no alphabetic
      characters and cannot be sanitized into a valid ID.
  """
  alphanums_hyphens = _NON_ALPHANUM_PATTERN.sub('-', flag_key.lower())
  stripped_leading = _LEADING_NON_CHAR_PATTERN.sub('', alphanums_hyphens)
  flag_id = stripped_leading[:63].rstrip('-')

  if not flag_id:
    raise exceptions.InvalidDataError(
        f'Flag {flag_key!r} invalid. Flag key must contain at least one letter.'
    )
  return flag_id


def _CreateFlag(
    flags_service: ezclient.SaasservicemgmtV1beta1.ProjectsLocationsFlagsService,
    parent: str,
    flag_msg: ezmessages.Flag,
    validate_only: bool = False,
) -> None:
  """Creates a flag in the App Lifecycle Manager API."""
  flag_id = _SanitizeFlagId(flag_msg.key)
  flags_create_req = (
      ezmessages.SaasservicemgmtProjectsLocationsFlagsCreateRequest(
          parent=parent,
          flagId=flag_id,
          validateOnly=validate_only,
          flag=flag_msg,
      )
  )
  flags_create_req_dict = encoding.MessageToDict(flags_create_req)
  if not validate_only:
    log.status.Print(
        f'\nAttempting to create flag:\n{pprint.pformat(flags_create_req_dict)}'
    )

  try:
    # Note: base_api client has a retry mechanism built in.
    flag_resp = flags_service.Create(flags_create_req)

    if not validate_only:
      log.status.Print(
          '\nCreated flag:\n'
          f'{pprint.pformat(encoding.MessageToDict(flag_resp))}'
      )

  except exceptions.HttpError as e:
    if _IsFlagAlreadyExistsError(e):
      existing_flag = _GetExistingFlag(flags_service, parent, flag_id)
      if existing_flag is None:
        raise

      log.debug(
          f'\nFlag {flag_id!r} already exists. Existing'
          f' flag:\n{pprint.pformat(encoding.MessageToDict(existing_flag))}'
      )

      if existing_flag.flagValueType != flag_msg.flagValueType:
        raise exceptions.InvalidDataError(
            f'Flag {flag_id!r} of type {flag_msg.flagValueType!r} already'
            f' exists with different value type {existing_flag.flagValueType}'
        ) from e
      else:
        raise exceptions.HttpConflictError(
            e.response, e.content, e.url, e.method_config, e.request
        )
    else:
      raise e


def _LogHttpError(
    e: exceptions.HttpError,
    flag_key: str,
    log_func: Callable[[str], None],
    msg_prefix: str = '',
):
  """Logs an HTTP error."""
  try:
    error_content = json.loads(e.content)
    error_message = error_content.get('error', {}).get(
        'message', 'Unknown error'
    )
    error_code = error_content.get('error', {}).get('code', 'Unknown error')
    log_func(f'{msg_prefix} {flag_key}: {error_code} : {error_message}')
  except json.JSONDecodeError:
    log_func(f'{msg_prefix} {flag_key}: {e.content}')


@base.DefaultUniverseOnly
@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA)
class Push(base.Command):
  """Create App Lifecycle Manager Flags from a JSON manifest file."""

  detailed_help = {
      'brief': 'Create Flags from a JSON manifest file.',
      'DESCRIPTION': """\
          Creates Flags based on the contents of a provided JSON manifest file. Flags defined in the file
          will be created. If a Flag already exists with the same type, it will be ignored. Otherwise,
          the command will exit with an error.
      """,
      'EXAMPLES': """\
          To create flags from a manifest file named `my_flags.json`:

              $ {command} --file=my_flags.json --unit-kind=my-unit-kind
          """,
  }

  @staticmethod
  def Args(parser):
    """Register flags for gcloud app-lifecycle-manager flags push: --file, --unit-kind."""
    parser.add_argument(
        '--file',
        type=arg_parsers.FileContents(),
        required=True,
        help=(
            'Path to the JSON manifest file containing the flag definitions.'
            ' The manifest file must conform to the OpenFeature CLI flag'
            ' manifest schema.'
        ),
    )

    # This also implicitly adds the --location flag.
    arg_utils.AddUnitKindArgToParser(
        parser,
        required=True,
        help_text='UnitKind to push all the flags in the manifest file to.',
    )

  def Run(self, args):
    """Execute the gcloud app-lifecycle-manager flags push command."""

    project = properties.VALUES.core.project.Get()
    unit_kind_ref = args.CONCEPTS.unit_kind.Parse()
    parent = unit_kind_ref.Parent().RelativeName()

    log.debug(f'--project={project}')
    log.debug(f'--location={args.location}')
    log.debug(f'--unit-kind={args.unit_kind}')
    log.debug(f'Parent reference: {unit_kind_ref.Parent()}')
    log.debug(f'Parent relative name: {parent}')
    log.debug(f'UnitKind Name: {unit_kind_ref.Name()}')
    log.debug(f'UnitKind RelativeName: {unit_kind_ref.RelativeName()}')

    client = saasservicemgmt_util.GetV1Beta1ClientInstance()
    flags_service = client.projects_locations_flags

    # Parse the manifest file.
    try:
      manifest = json.loads(args.file)
    except json.JSONDecodeError as e:
      raise calliope_exceptions.ToolException(
          f'Failed to parse JSON: {e}'
      ) from e

    # Validate the manifest file.
    try:
      flags = _GetFlagMessagesFromManifest(
          manifest, unit_kind_ref.RelativeName()
      )
    except exceptions.InvalidDataError as e:
      raise calliope_exceptions.ToolException(
          f'Failed to parse flags: {e}'
      ) from e

    # Validate the flags.
    log.status.Print(f'Found {len(flags)} flags in manifest. Validating...')
    try:
      valid_flags = _GetValidFlags(flags, flags_service, parent)
    except exceptions.InvalidDataError as e:
      raise calliope_exceptions.ToolException(
          f'No flags were created: {e}'
      ) from e

    # Create the flags.
    if not valid_flags:
      log.status.Print('No valid flags to create. No flags were created.')
      return

    log.status.Print(
        'All flags validated successfully. Creating'
        f' {len(valid_flags)} flags...'
    )

    for flag_msg in valid_flags:
      try:
        _CreateFlag(flags_service, parent, flag_msg, validate_only=False)
      except exceptions.HttpError as e:
        _LogHttpError(e, flag_msg.key, log.error, 'Failed to create flag')
        raise calliope_exceptions.ToolException(
            f'Halting execution due to post-validation error: {e}'
        ) from e

    log.status.Print(f'{len(valid_flags)} flags created successfully.')
